@@ -5,13 +5,13 @@ __all__ = [
 
 from dataclasses import InitVar, dataclass, field
 from pathlib import Path
-from typing import Any, Dict, Literal, MutableMapping, Optional, Tuple, Union, overload
+from typing import Any, Dict, List, Literal, Optional, Tuple, Union, overload
 
 from beet import Cache, Context, DataPack, File, ResourcePack
 from beet.core.utils import FileSystemPath, extra_field
 
-from .directive import Directive, get_builtin_directives
-from .extract import MarkdownExtractor, TextExtractor
+from .directive import DirectiveRegistry
+from .extract import FragmentLoader, MarkdownExtractor, TextExtractor
 from .serialize import ExternalFilesManager, MarkdownSerializer, TextSerializer
 
 
@@ -29,9 +29,8 @@ class Document:
     assets: ResourcePack = field(default_factory=ResourcePack)
     data: DataPack = field(default_factory=DataPack)
 
-    directives: MutableMapping[str, Directive] = extra_field(
-        default_factory=get_builtin_directives
-    )
+    loaders: List[FragmentLoader] = extra_field(default_factory=list)
+    directives: DirectiveRegistry = extra_field(default_factory=DirectiveRegistry)
 
     text_extractor: TextExtractor = extra_field(default_factory=TextExtractor)
     markdown_extractor: MarkdownExtractor = extra_field(
@@ -57,9 +56,14 @@ class Document:
             self.data = ctx.data
             if cache is None:
                 cache = ctx.cache["lectern"]
+
         if cache:
             self.text_extractor.cache = cache
             self.markdown_extractor.cache = cache
+
+        self.directives.assets = self.assets
+        self.directives.data = self.data
+
         if path:
             self.load(path)
         if text:
@@ -77,7 +81,11 @@ class Document:
 
     def add_text(self, source: str):
         """Extract pack fragments from plain text."""
-        assets, data = self.text_extractor.extract(source, self.directives)
+        assets, data = self.text_extractor.extract(
+            source=source,
+            directives=self.directives.resolve(),
+            loaders=self.loaders,
+        )
         self.assets.merge(assets)
         self.data.merge(data)
 
@@ -89,7 +97,8 @@ class Document:
         """Extract pack fragments from markdown."""
         assets, data = self.markdown_extractor.extract(
             source=source,
-            directives=self.directives,
+            directives=self.directives.resolve(),
+            loaders=self.loaders,
             external_files=external_files,
         )
         self.assets.merge(assets)
@@ -97,7 +106,11 @@ class Document:
 
     def get_text(self) -> str:
         """Turn the data pack and the resource pack into text."""
-        return self.text_serializer.serialize(self.assets, self.data)
+        return self.text_serializer.serialize(
+            assets=self.assets,
+            data=self.data,
+            mapping=self.directives.resolve().get_serialization_mapping(),
+        )
 
     @overload
     def get_markdown(
@@ -124,6 +137,7 @@ class Document:
         content = self.markdown_serializer.serialize(
             assets=self.assets,
             data=self.data,
+            mapping=self.directives.resolve().get_serialization_mapping(),
             external_files=external_files,
             external_prefix=prefix,
         )
